#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>

typedef unsigned char ubyte;

#include "key.c"

char *program_name;

#define FRMT_2FILES 1
#define FRMT_RAW 2

void error(char *msg) {
  fprintf(stderr, "%s: %s\n", program_name, msg);
}

void error_sys(char *fname) {
  fprintf(stderr, "%s: %s: %s\n", program_name, fname, strerror(errno));
}

int decode_block(ubyte *in, ubyte *out) {
  int i, j;
  for (i = 0; i < 5; i++)
    for (j = 0; j < 50; j++)
      out[i * 50 + j] = key[i * 50 + j][in[i * 51 + j]];
  return 0;
}

void load_key(void) {
  FILE *fp;
  int i;
  fp = fopen("key.bin", "rb");
  if (!fp) {
    error("cannot open key");
    exit(1);
  }

  for (i = 0; i < 250; i++)
    fread(key[i], 1, 256 ,fp);
  fclose(fp);
}

void usage(FILE *fp) {
  fprintf(fp, "usage: %s [-f format] input_file output_file ...\n"
	  "format is one of raw (default), 2files\n",
	  program_name);
}

int decode_2files(char *name, char *out_name1, char *out_name2) {
  FILE *fp, *out_fp;
  ubyte in_buf[512];
  ubyte out_buf[250];
  int len, len1, check_sum;
  int i, j;
  int sum;
  char str[20];

  fp = fopen(name, "rb");
  if (!fp) {
    error_sys(name);
    return -1;
  }

  out_fp = fopen(out_name1, "wb");
  if (!out_fp) {
    error_sys(out_name1);
    fclose(fp);
    return -1;
  }

  if (fread(in_buf, 1, 512, fp) != 512) {
    fclose(out_fp);
    fclose(fp);
    return -2;
  }

  if (decode_block(in_buf + 256, out_buf)) {
    fclose(out_fp);
    fclose(fp);
    return -3;
  }

  if (decode_block(in_buf, out_buf)) {
    fclose(out_fp);
    fclose(fp);
    return -3;
  }

  strncpy(str, out_buf, sizeof(str) - 1);
  str[sizeof(str) - 1] = 0;
  printf("name: %s\n", str, out_buf[0], out_buf[1], out_buf[2]);

  len = (out_buf[0x10] << 24) | (out_buf[0x20] << 16)
    | (out_buf[0x30] << 8) | out_buf[0x40];
  len1 = (out_buf[0x50] << 24) | (out_buf[0x60] << 16)
    | (out_buf[0x70] << 8) | out_buf[0x80];
  check_sum = (out_buf[0x90] << 24) | (out_buf[0xa0] << 16)
    | (out_buf[0xb0] << 8) | out_buf[0xc0];

  printf("len: %08x, len1: %08x\n", len, len1);

  sum = 0;
  for (i = 0x200; i < len1; i += 0x100) {
    if (fread(in_buf, 1, 256, fp) != 256) {
      fclose(out_fp);
      fclose(fp);
      return -2;
    }

    for (j = 0; j < 256; j++)
      sum += in_buf[j];

    if (decode_block(in_buf, out_buf)) {
      fclose(out_fp);
      fclose(fp);
      return -3;
    }
    if (fwrite(out_buf, 1, 250, out_fp) != 250) {
      error_sys(out_name2);
      fclose(out_fp);
      fclose(fp);
      return -2;
    }
  }

  if (fclose(out_fp)) {
    error_sys(out_name1);
    return -2;
  }

  out_fp = fopen(out_name2, "wb");
  if (!out_fp) {
    error_sys(out_name2);
    fclose(fp);
    return -1;
  }

  for (; i < len; i += 0x100) {
    if (fread(in_buf, 1, 256, fp) != 256) {
      fclose(out_fp);
      fclose(fp);
      return -2;
    }

    for (j = 0; j < 256; j++)
      sum += in_buf[j];

    if (decode_block(in_buf, out_buf)) {
      fclose(out_fp);
      fclose(fp);
      return -3;
    }
    if (fwrite(out_buf, 1, 250, out_fp) != 250) {
      error_sys(out_name2);
      fclose(out_fp);
      fclose(fp);
      return -2;
    }
  }

  fclose(fp);
  if (fclose(out_fp)) {
    error_sys(out_name2);
    return -2;
  }

  if (sum != check_sum)
    return -4;

  return 0;
}

int decode_raw(char *in_name, char *out_name) {
  FILE *fp, *out_fp;
  ubyte in_buf[256], out_buf[250];

  fp = fopen(in_name, "rb");
  if (!fp) {
    error_sys(in_name);
    return -1;
  }

  out_fp = fopen(out_name, "wb");
  if (!out_fp) {
    error_sys(out_name);
    fclose(fp);
    return -1;
  }

  while (fread(in_buf, 1, 256, fp) == 256) {
    if (decode_block(in_buf, out_buf)) {
      fclose(out_fp);
      fclose(fp);
      return -3;
    }
    if (fwrite(out_buf, 1, 250, out_fp) != 250) {
      error_sys(out_name);
      fclose(out_fp);
      fclose(fp);
      return -2;
    }
  }

  fclose(fp);
  if (fclose(out_fp)) {
    error_sys(out_name);
    return -2;
  }
    
  return 0;
}

int main(int argc, char *argv[]) {
  char *inname, *outname[2];
  char tmp[200];
  int format, r, c, i;

  program_name = argv[0];

  format = FRMT_RAW;

  while ((c = getopt(argc, argv, "f:")) != -1)
    switch (c) {
    case 'f':
      if (!strcmp(optarg, "2files"))
	format = FRMT_2FILES;
      else if (!strcmp(optarg, "raw"))
	format = FRMT_RAW;
      else {
	snprintf(tmp, sizeof(tmp), "unknown format: %s", optarg);
	error(tmp);
	usage(stderr);
	return 1;
      }
      break;
    case '?':
      if (isprint (optopt))
	snprintf(tmp, sizeof(tmp), "unknown option `-%c'.\n", optopt);
      else
	snprintf(tmp, sizeof(tmp), "unknown option character `\\x%x'.\n",
		 optopt);
      error(tmp);
      usage(stderr);
      return 1;
    default:
      abort();
    }


  if (optind < argc)
    inname = argv[optind];
  else {
    error("no input file specified");
    usage(stderr);
    return 1;
  }


  for (i = 0; i < (format == FRMT_2FILES ? 2 : 1); i++)
    if (optind + 1 + i < argc)
      outname[i] = argv[optind + 1 + i];
    else {
      error("no output file specified");
      usage(stderr);
      return 1;
    }

  switch (format) {
  case FRMT_2FILES:
    r = decode_2files(inname, outname[0], outname[1]);
    break;
  case FRMT_RAW:
    r = decode_raw(inname, outname[0]);
    break;
  }

  if (r) {
    error("error during decoding");
    return 1;
  }

  return 0;
}
